import greenfoot.World;
import greenfoot.MouseInfo;
import greenfoot.Greenfoot;
import greenfoot.GreenfootImage;
import java.util.ArrayList;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Color;

/**
 * Screen
 * 
 * @author Taylor Born
 * @version May 2013 - June 2013
 */
public class Screen extends World
{
    public static Screen main;
    private static final int WORLD_WIDTH = 600;
    private static final int WORLD_HEIGHT = 425;

    private final GreenfootImage PIC = new GreenfootImage(WORLD_WIDTH, WORLD_HEIGHT);
    private final GreenfootImage LAYER_SHADOWS = new GreenfootImage(WORLD_WIDTH, WORLD_HEIGHT);
    private final Graphics2D LAYER_SHADOWS_GRAPHICS = (Graphics2D)LAYER_SHADOWS.getAwtImage().getGraphics();
    private final GreenfootImage LAYER_2 = new GreenfootImage(WORLD_WIDTH, WORLD_HEIGHT);
    private final GreenfootImage BALL = new GreenfootImage("Ball.png");
    private final GreenfootImage BALL_SHADOW = new GreenfootImage("Ball Shadow.png");
    
    private final Color[] BLOCK_COLORS = { Color.RED, Color.GRAY, Color.YELLOW, Color.BLUE, Color.GREEN };
    private final GreenfootImage[][] ISO_BLOCKS = new GreenfootImage[BLOCK_COLORS.length][5];
    
    private final GreenfootImage ISO_BLOCK_SHADOW = new GreenfootImage("Block Shadow.png");
    private final GreenfootImage ISO_BLOCK_SHADOW2 = new GreenfootImage(ISO_BLOCK_SHADOW);
    private final GreenfootImage WINDOW_VERTICAL = new GreenfootImage("Window.png");
    private final GreenfootImage WINDOW_HORIZONTAL = new GreenfootImage(WINDOW_VERTICAL);
    private final GreenfootImage MISSILE = new GreenfootImage("Missile.png");
    private final GreenfootImage MISSILE_SHADOW = new GreenfootImage("MissileShadow.png");
    private final GreenfootImage PICKUP = new GreenfootImage("Power.png");
    
    private final SoundCluster BOUNCE_SOUND = new SoundCluster("Bounce.wav", 75);
    private final SoundSingle FINISH_SOUND = new SoundSingle("Finish.wav", 75);
    
    private final int SIZE = 12;
    private final Block[][] GRID = new Block[SIZE][SIZE];
    private final int GRID_BLOCK_SIDE_LENGTH = 30;
    private final int TOP_OFFSET = 45;
    private final int ISO_BASE_IMAGE_WIDTH = 49;
    private final int ISO_BASE_IMAGE_HEIGHT = 29;
    private final int X_OFFSET = (WORLD_WIDTH - (ISO_BASE_IMAGE_WIDTH - 1) * SIZE - 1) / 2;
    
    private final int ISO_HEIGHT = 25 - 1;
    private final double ISO_BLOCK_SIDE_LENGTH = Math.sqrt(Math.pow(ISO_BASE_IMAGE_WIDTH / 2, 2) + Math.pow(ISO_BASE_IMAGE_HEIGHT / 2, 2));
    private final double ANGLE = Math.atan2(ISO_BASE_IMAGE_HEIGHT / 2, ISO_BASE_IMAGE_WIDTH / 2);
    private final double SLOPE = 15d / 25;
    private double paddle;
    private int paddleScale;
    private final int NORMAL_SCALE = 2;
    private final int MAX_SCALE = 3;
    private final int MIN_SCALE = 1;
    private final GreenfootImage[] PADDLE_IMAGES = new GreenfootImage[MAX_SCALE];
    private final int BALL_SIZE = 6;
    private final ArrayList<Ball> BALLS = new ArrayList<Ball>();
    private int speed;
    private final int NORMAL_SPEED = 3;
    private final int MAX_SPEED = 5;
    private final int MIN_SPEED = 2;
    private int blockCount;
    private Point lastMouse = new Point(0, 0);
    private int level;
    private int lives;
    private int currentPower;
    private boolean barrier;
    private final int[] BARRIER_CLIP_X = { X_OFFSET, X_OFFSET, 0, 0, 0, 0, X_OFFSET, X_OFFSET };
    private final int BB = (TOP_OFFSET + (SIZE - 1) * (ISO_BASE_IMAGE_HEIGHT / 2)) + ISO_BASE_IMAGE_HEIGHT / 2;
//     private final int[] BARRIER_CLIP_Y = { BB, BB + ISO_HEIGHT / 6, 0, 0, 0, 0, BB + ISO_HEIGHT / 6 + (int)(Math.sin(ANGLE) * ISO_BLOCK_SIDE_LENGTH * SIZE), BB + (int)(Math.sin(ANGLE) * ISO_BLOCK_SIDE_LENGTH * SIZE) };
    private final ArrayList<Missile> MISSILES = new ArrayList<Missile>();
    private final ArrayList<Pickup> PICKUPS = new ArrayList<Pickup>();
    
    private static final int CODE_ID = 72;

    public Screen()
    {
        super(WORLD_WIDTH, WORLD_HEIGHT, 1);
        main = this;
        CodeKeyUser.enter();
        initializeImages();
        newGame();
        draw();
    }
    private void initializeImages()
    {
        setBackground(PIC);
        LAYER_SHADOWS.setTransparency(200);
        LAYER_SHADOWS_GRAPHICS.setColor(new Color(0, 0, 0, 0));
        WINDOW_HORIZONTAL.mirrorHorizontally();
        
        for (int i = MIN_SCALE; i < MAX_SCALE + 1; i++)
            PADDLE_IMAGES[i - MIN_SCALE] = new GreenfootImage("Paddle" + i + ".png");
        
        ISO_BLOCKS[0][0] = new GreenfootImage("RedSide.png"); // Side Left
        ISO_BLOCKS[0][1] = new GreenfootImage(ISO_BLOCKS[0][0]); // Side Right
        ISO_BLOCKS[0][2] = new GreenfootImage("RedBase.png"); // Base Left
        ISO_BLOCKS[0][3] = new GreenfootImage(ISO_BLOCKS[0][2]); // Base Right
        ISO_BLOCKS[0][4] = new GreenfootImage("RedEnd.png"); // End
        ISO_BLOCKS[0][1].rotate(180); // Side right
        ISO_BLOCKS[0][3].rotate(180); // Base right
        
        final Color TRANSPARENT = new Color(0, 0, 0, 0);
        
        for (int i = 0; i < ISO_BLOCKS[0][1].getHeight(); i++)
        {
            ISO_BLOCKS[0][1].setColorAt(0, i, TRANSPARENT); // Side right
            ISO_BLOCKS[0][4].setColorAt(0, i, TRANSPARENT); // End
        }
        // Base Right
        for (int i = 0; i < ISO_BLOCKS[0][3].getWidth(); i++)
            for (int k = ISO_BLOCKS[0][3].getHeight() - 1; true; k--)
                if (ISO_BLOCKS[0][3].getColorAt(i, k).getAlpha() != 0)
                {
                    ISO_BLOCKS[0][3].setColorAt(i, k, TRANSPARENT);
                    break;
                }
        // Base Left
        for (int i = 0; i < ISO_BLOCKS[0][2].getWidth(); i++)
            for (int k = ISO_BLOCKS[0][2].getHeight() - 1; true; k--)
                if (ISO_BLOCKS[0][2].getColorAt(i, k).getAlpha() != 0)
                {
                    ISO_BLOCKS[0][2].setColorAt(i, k, TRANSPARENT);
                    break;
                }
        ISO_BLOCKS[0][2].setColorAt(42, 17, TRANSPARENT);
        ISO_BLOCKS[0][2].setColorAt(30, 24, TRANSPARENT);
        
        for (int i = 1; i < BLOCK_COLORS.length; i++)
            for (int k = 0; k < ISO_BLOCKS[0].length; k++)
            {
                ISO_BLOCKS[i][k] = new GreenfootImage(ISO_BLOCKS[0][k]);
                for (int y = 0; y < ISO_BLOCKS[i][k].getHeight(); y++)
                    for (int x = 0; x < ISO_BLOCKS[i][k].getWidth(); x++)
                    if (ISO_BLOCKS[i][k].getColorAt(x, y).getAlpha() != 0)
                        {
                            double r = ISO_BLOCKS[i][k].getColorAt(x, y).getRed() / 255d;
                            ISO_BLOCKS[i][k].setColorAt(x, y, new Color((int)(BLOCK_COLORS[i].getRed() * r), (int)(BLOCK_COLORS[i].getGreen() * r), (int)(BLOCK_COLORS[i].getBlue() * r)));
                        }
            }
//         for (int k = 0; k < ISO_BLOCKS[0].length; k++)
//             try
//             {
//                 javax.imageio.ImageIO.write(ISO_BLOCKS[0][k].getAwtImage(), "png", new java.io.File("images/New/" + k + "" + ".png"));
//             }
//             catch (java.io.IOException ex)
//             {}
    }
    private void newGame()
    {
        level = 0;
        newMap(level);
        lives = 3;
    }
    private void newMap(int map)
    {
        MISSILES.clear();
        PICKUPS.clear();
        currentPower = Power.NULL_POWER;
        barrier = false;
        paddleScale = NORMAL_SCALE;
        
        blockCount = 0;
        speed = NORMAL_SPEED;
        BALLS.clear();
        BALLS.add(new Ball());
        
        for (int i = 0; i < SIZE; i++)
            for (int k = 0; k < SIZE; k++)
                if (i >= MAPS[map].length)
                    GRID[i][k] = new Block(k, i, true);
                else
                {
                    char c = MAPS[map][i].charAt(k / 2);
                    GRID[i][k] = newBlock(c, k, i);
                    GRID[i][k + 1] = newBlock(c, ++k, i);
                    if (c != '_' && c != '1')
                        blockCount++;
                }
    }
    private Block newBlock(int type, int x, int y)
    {
        switch (type)
        {
            case '0': return new Block(x, y);
            case '1': return new SteelBlock(x, y);
            case '2': return new MultiBlock(x, y);
            case '3': return new MissileBlock(x, y);
            case '4': return new BombBlock(x, y);
        }
        return new Block(x, y, true);
    }
    public void act()
    {
        handleManualProgression(); // For testing purposes
        updateLastMouse();
        actMissiles();
        actPickups();
        actBalls();
        draw();
    }
    private void handleManualProgression()
    {
        String key = Greenfoot.getKey();
        if (key != null)
            if (key.equals("left") || key.equals("a"))
            {
                if (level > 0)
                    newMap(--level);
            }
            else if (key.equals("right") || key.equals("d"))
                if (level < MAPS.length - 1)
                    newMap(++level);
    }
    private void updateLastMouse()
    {
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if (Greenfoot.mouseMoved(null) || Greenfoot.mouseDragged(null))
            lastMouse = new Point(mouse.getX(), mouse.getY());
    }
    private void actMissiles()
    {
        for (int m = 0; m < MISSILES.size(); m++)
        {
            Missile missile = MISSILES.get(m);
            missile.act();
            if (missile.getY() > SIZE * GRID_BLOCK_SIDE_LENGTH)
            {
                MISSILES.remove(m--);
                if (missile.getX() > paddle && missile.getX() < paddle + GRID_BLOCK_SIDE_LENGTH * paddleScale)
                    lose();
            }
        }
    }
    private void actPickups()
    {
        for (int p = 0; p < PICKUPS.size(); p++)
        {
            Pickup pickup = PICKUPS.get(p);
            pickup.act();
            if (pickup.getY() > SIZE * GRID_BLOCK_SIDE_LENGTH)
            {
                PICKUPS.remove(p--);
                if (pickup.getX() > paddle && pickup.getX() < paddle + GRID_BLOCK_SIDE_LENGTH * paddleScale)
                    pickup.apply();
            }
        }
    }
    private void actBalls()
    {
        boolean clicked = Greenfoot.mouseClicked(this);
        
        for (int b = 0; b < BALLS.size(); b++)
            if (BALLS.get(b).act())
            {
                BALLS.remove(b--);
                if (BALLS.isEmpty())
                    lose();
            }
            else if (clicked)
                BALLS.get(b).release();
    }
    private void draw()
    {
        clearLayers();
        drawBlocks();
        drawPaddle();
        PIC.drawImage(LAYER_SHADOWS, 0, 0);
        PIC.drawImage(LAYER_2, 0, 0);
    }
    private void clearLayers()
    {
        PIC.clear();
        PIC.setColor(Color.WHITE);
        PIC.fill();
        LAYER_SHADOWS.clear();
        LAYER_2.clear();
    }
    private void drawBlocks()
    {
        for (int y = 0; y < SIZE; y++)
            for (int h = y; h > -1; h--) // Reading diagonal (southwest to northeast) for top half of map.
                drawBlock(GRID[h][y - h], y - h, y, X_OFFSET + ISO_BASE_IMAGE_WIDTH / 2 * (SIZE - 1 - y));
        for (int x = 1; x < SIZE; x++)
            for (int v = x; v < SIZE; v++)
                drawBlock(GRID[SIZE - (v - x + 1)][v], v - x, SIZE + x - 1, X_OFFSET + ISO_BASE_IMAGE_WIDTH / 2 * x);
    }
    private void drawBlock(Block block, int x, int y, int xOffSet)
    {
        // Top Boundary
        if (block.getY() == 0)
            for (int i = 0; i < 3; i++)
                LAYER_2.drawImage(WINDOW_HORIZONTAL, xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1) + ISO_BASE_IMAGE_WIDTH / 2, TOP_OFFSET + y * (ISO_BASE_IMAGE_HEIGHT / 2) - ISO_HEIGHT * 3 / 2 + ISO_HEIGHT * i);
        // Left Boundary
        if (block.getX() == 0)
            for (int i = 0; i < 3; i++)
                LAYER_2.drawImage(WINDOW_VERTICAL, xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1), TOP_OFFSET + y * (ISO_BASE_IMAGE_HEIGHT / 2) - ISO_HEIGHT * 3 / 2 + ISO_HEIGHT * i);
        
        // Missiles passing through
        for (Missile missile : MISSILES)
            if (block.getX() == missile.getX() / GRID_BLOCK_SIDE_LENGTH && block.getY() == missile.getY() / GRID_BLOCK_SIDE_LENGTH)
            {
                int missile_loc_x = X_OFFSET + 24 * (SIZE - 1) + ISO_BASE_IMAGE_WIDTH / 2 + (int)(Math.cos(ANGLE) * ((double)missile.getX() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH) + Math.cos(ANGLE + (Math.PI - ANGLE * 2)) * ((double)missile.getY() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH));
                int missile_loc_y = TOP_OFFSET + (int)(Math.sin(ANGLE) * ((double)missile.getX() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH) + Math.sin(ANGLE + (Math.PI - ANGLE * 2)) * ((double)missile.getY() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH)) - missile.getHeight();
                LAYER_2.drawImage(MISSILE, missile_loc_x - MISSILE.getWidth() / 2, missile_loc_y - MISSILE.getHeight() / 2);
                LAYER_SHADOWS_GRAPHICS.drawImage(MISSILE_SHADOW.getAwtImage(), null, missile_loc_x - MISSILE_SHADOW.getWidth() / 2, missile_loc_y + missile.getHeight() - MISSILE_SHADOW.getHeight() / 2 + ISO_HEIGHT * 3 / 2);
            }
        for (Pickup pickup : PICKUPS)
            if (block.getX() == pickup.getX() / GRID_BLOCK_SIDE_LENGTH && block.getY() == pickup.getY() / GRID_BLOCK_SIDE_LENGTH)
            {
                int pickup_loc_x = X_OFFSET + 24 * (SIZE - 1) + ISO_BASE_IMAGE_WIDTH / 2 + (int)(Math.cos(ANGLE) * ((double)pickup.getX() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH) + Math.cos(ANGLE + (Math.PI - ANGLE * 2)) * ((double)pickup.getY() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH));
                int pickup_loc_y = TOP_OFFSET + (int)(Math.sin(ANGLE) * ((double)pickup.getX() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH) + Math.sin(ANGLE + (Math.PI - ANGLE * 2)) * ((double)pickup.getY() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH));
                LAYER_2.drawImage(PICKUP, pickup_loc_x - PICKUP.getWidth() / 2, pickup_loc_y - PICKUP.getHeight() / 2);
            }
        
        drawBalls(block.getX(), block.getY());
        
        int y_loc = TOP_OFFSET + y * (ISO_BASE_IMAGE_HEIGHT / 2) - ISO_HEIGHT / 2;
        
        // The Block
        if (!block.isDestroyed() || block.isFading())
        {
            int tt = block.isFading() ? block.getTransparency() : 255;
            int t = tt;
            if (block instanceof MultiBlock)
            {
                int h = ((MultiBlock)block).getHeight();
                y_loc += h;
                
                drawImageOnImage(LAYER_2, xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1), y_loc + ISO_HEIGHT, block.getX() % 2 == 0 ? getIsoImageBaseLeft(block.getWallType()) : getIsoImageBaseRight(block.getWallType()), t);
                // Block-side
                drawImageOnImage(LAYER_2, xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1), y_loc + ISO_BASE_IMAGE_HEIGHT / 2 + ISO_HEIGHT, block.getX() % 2 == 0 ? getIsoImageSideLeft(block.getWallType()) : getIsoImageSideRight(block.getWallType()), t);
                
                if (block.getX() % 2 == 1)
                {
                    // Block-end-side
                    drawImageOnImage(LAYER_2, xOffSet + x * 48 + ISO_BASE_IMAGE_WIDTH / 2, y_loc + ISO_BASE_IMAGE_HEIGHT / 2 + ISO_HEIGHT, getIsoImageEnd(block.getWallType()), t);
                }
                t = (int)((1 - -h / 24d) * 255);
            }
            
            // Block-base
            drawImageOnImage(LAYER_2, xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1), y_loc, block.getX() % 2 == 0 ? getIsoImageBaseLeft(block.getWallType()) : getIsoImageBaseRight(block.getWallType()), t);
            // Block-side
            drawImageOnImage(LAYER_2, xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1), y_loc + ISO_BASE_IMAGE_HEIGHT / 2, block.getX() % 2 == 0 ? getIsoImageSideLeft(block.getWallType()) : getIsoImageSideRight(block.getWallType()), t);
            
            if (block instanceof MissileBlock)
            {
                int missile_loc_x = xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1) + ISO_BASE_IMAGE_WIDTH / 2 + 1;
                int missile_loc_y = TOP_OFFSET + y * (ISO_BASE_IMAGE_HEIGHT / 2) - ISO_HEIGHT / 2 + ISO_BASE_IMAGE_HEIGHT / 2;
                LAYER_2.drawImage(MISSILE, missile_loc_x - MISSILE.getWidth() / 2, missile_loc_y - MISSILE.getHeight() / 2);
                LAYER_SHADOWS_GRAPHICS.drawImage(MISSILE_SHADOW.getAwtImage(), null, missile_loc_x - MISSILE_SHADOW.getWidth() / 2, missile_loc_y + ISO_HEIGHT / 2 - MISSILE_SHADOW.getHeight() / 2 + ISO_HEIGHT * 3 / 2);
            }
            
            if (block.getX() % 2 == 1)
            {
                // Block-end-side
                drawImageOnImage(LAYER_2, xOffSet + x * 48 + ISO_BASE_IMAGE_WIDTH / 2, y_loc + ISO_BASE_IMAGE_HEIGHT / 2, getIsoImageEnd(block.getWallType()), t);
            }
            else
            {
                // Block shadow
                ISO_BLOCK_SHADOW.setTransparency(tt);
                ISO_BLOCK_SHADOW2.clear();
                ISO_BLOCK_SHADOW2.drawImage(ISO_BLOCK_SHADOW, 0, 0);
                LAYER_SHADOWS_GRAPHICS.drawImage(ISO_BLOCK_SHADOW2.getAwtImage(), null, xOffSet + x * (ISO_BASE_IMAGE_WIDTH - 1) - 5, TOP_OFFSET + y * (ISO_BASE_IMAGE_HEIGHT / 2) + ISO_HEIGHT * 3 / 2 - 5);
            }
        }
        
        // Right Boundary
        if (block.getX() == SIZE - 1)
            for (int i = 0; i < 3; i++)
                LAYER_2.drawImage(WINDOW_VERTICAL, xOffSet + x * 48 + ISO_BASE_IMAGE_WIDTH / 2, TOP_OFFSET + y * (ISO_BASE_IMAGE_HEIGHT / 2) + ISO_BASE_IMAGE_HEIGHT / 2 - ISO_HEIGHT * 3 / 2 + ISO_HEIGHT * i);
    }
    private void drawImageOnImage(GreenfootImage canvas, int x, int y, GreenfootImage pic, int transparency)
    {
        pic.setTransparency(transparency);
        canvas.drawImage(pic, x, y);
    }
    private void drawBalls(int blockX, int blockY)
    {
        for (Ball ball : BALLS)
            if (blockX == (int)ball.getX() / GRID_BLOCK_SIDE_LENGTH && blockY == (int)ball.getY() / GRID_BLOCK_SIDE_LENGTH)
            {
                int ball_loc_x = X_OFFSET + 24 * (SIZE - 1) + ISO_BASE_IMAGE_WIDTH / 2 + (int)(Math.cos(ANGLE) * (ball.getX() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH) + Math.cos(ANGLE + (Math.PI - ANGLE * 2)) * (ball.getY() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH));
                int ball_loc_y = TOP_OFFSET + (int)(Math.sin(ANGLE) * (ball.getX() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH) + Math.sin(ANGLE + (Math.PI - ANGLE * 2)) * (ball.getY() / GRID_BLOCK_SIDE_LENGTH * ISO_BLOCK_SIDE_LENGTH));
                LAYER_2.drawImage(BALL, ball_loc_x - BALL_SIZE / 2, ball_loc_y - BALL_SIZE / 2);
                LAYER_2.drawImage(BALL_SHADOW, ball_loc_x - BALL_SIZE / 2, ball_loc_y - BALL_SIZE / 2 + ISO_HEIGHT * 3 / 2);
            }
    }
    private void drawPaddle()
    {
        int x = X_OFFSET;
        int y = (TOP_OFFSET + (SIZE - 1) * (ISO_BASE_IMAGE_HEIGHT / 2)) + ISO_BASE_IMAGE_HEIGHT / 2;
        
        double b = y - SLOPE * x;
        
        double intersectX = (b - (lastMouse.getY() - (- SLOPE) * lastMouse.getX())) / ((- SLOPE) - SLOPE);
        double intersectY = SLOPE * intersectX + b;
        
        int r = (int)Math.ceil(paddleScale * ISO_BLOCK_SIDE_LENGTH / 2);
        
        double distance;
        if (intersectX <= x)
            distance = r;
        else
        {
            distance = Math.sqrt(Math.pow(x - intersectX, 2) + Math.pow(y - intersectY, 2));
            
            if (distance < r)
                distance = r;
            else if (distance + r > ISO_BLOCK_SIDE_LENGTH * SIZE)
                distance = ISO_BLOCK_SIDE_LENGTH * SIZE - r;
        }
        
        paddle = (distance - r) / ISO_BLOCK_SIDE_LENGTH * GRID_BLOCK_SIDE_LENGTH;
        
        if (barrier)
        {
            int[][] xs = new int[2][2];
            int[][] ys = new int[2][2];
            
            // Left side.
            xs[0][0] = x;
            ys[0][0] = y - ISO_HEIGHT / 6;
            xs[0][1] = xs[0][0] + (int)(Math.cos(ANGLE) * (distance - r));
            ys[0][1] = ys[0][0] + (int)(Math.sin(ANGLE) * (distance - r));
            
            // Right side.
            xs[1][0] = xs[0][0] + (int)(Math.cos(ANGLE) * (distance + r));
            ys[1][0] = ys[0][0] + (int)(Math.sin(ANGLE) * (distance + r));
            xs[1][1] = xs[0][0] + (int)(Math.cos(ANGLE) * ISO_BLOCK_SIDE_LENGTH * SIZE);
            ys[1][1] = ys[0][0] + (int)(Math.sin(ANGLE) * ISO_BLOCK_SIDE_LENGTH * SIZE);
            
            Graphics2D g = (Graphics2D)LAYER_2.getAwtImage().getGraphics();
            if (++barrierStep == 80)
                barrierStep = 0;
            for (int h = 0; h < ISO_HEIGHT / 3; h++)
            {
                int d = (h + barrierStep / 10) % 8;
                if (d > 3)
                    d = 8 - d;
                int[] xx = {xs[0][0], xs[0][0], xs[0][1], xs[0][1], xs[1][0], xs[1][0], xs[1][1], xs[1][1]};
                int[] yy = {ys[0][0], ys[0][0] + ISO_HEIGHT / 3, ys[0][1] + ISO_HEIGHT / 3, ys[0][1], ys[1][0], ys[1][0] + ISO_HEIGHT / 3, ys[1][1] + ISO_HEIGHT / 3, ys[1][1]};
                g.setClip(new java.awt.Polygon(xx, yy, 8));
                
                g.setColor(new Color(0, 0, 255, 10 + (int)(d / 4d * 170)));
                g.drawLine(xs[0][0], ys[0][0] + h, xs[1][1], ys[1][1] + h);
            }
            g.dispose();
        }
        
        LAYER_2.drawImage(PADDLE_IMAGES[paddleScale - 1], x + (int)(Math.cos(ANGLE) * distance) - PADDLE_IMAGES[paddleScale - 1].getWidth() / 2, y + (int)(Math.sin(ANGLE) * distance) - PADDLE_IMAGES[paddleScale - 1].getHeight() / 2);
    }
    private int barrierStep;
    private GreenfootImage getIsoImageSideLeft(int i)
    {
        return ISO_BLOCKS[i][0];
    }
    private GreenfootImage getIsoImageSideRight(int i)
    {
        return ISO_BLOCKS[i][1];
    }
    private GreenfootImage getIsoImageBaseLeft(int i)
    {
        return ISO_BLOCKS[i][2];
    }
    private GreenfootImage getIsoImageBaseRight(int i)
    {
        return ISO_BLOCKS[i][3];
    }
    private GreenfootImage getIsoImageEnd(int i)
    {
        return ISO_BLOCKS[i][4];
    }
    
    private void lose()
    {
        BALLS.clear();
        BALLS.add(new Ball());
        currentPower = Power.NULL_POWER;
        lives--;
    }
    
    /**
     * 
     */
    public void newMissile(int x, int y)
    {
        MISSILES.add(new Missile(x * GRID_BLOCK_SIDE_LENGTH + GRID_BLOCK_SIDE_LENGTH / 2, y * GRID_BLOCK_SIDE_LENGTH + GRID_BLOCK_SIDE_LENGTH / 2));
    }
    
    public void addLife()
    {
        lives++;
    }
    public void speedUp()
    {
        if (speed < MAX_SPEED)
            updateSpeed(1);
    }
    public void speedDown()
    {
        if (speed > MIN_SPEED)
            updateSpeed(-1);
    }
    private void updateSpeed(int s)
    {
        speed += s;
        for (Ball ball : BALLS)
            ball.updateSpeed();
    }
    public void addBarrier()
    {
        barrier = true;
    }
    public void splitBall()
    {
        BALLS.get(0).split();
    }
    public void increasePaddle()
    {
        if (paddleScale < MAX_SCALE)
            paddleScale++;
    }
    public void decreasePaddle()
    {
        if (paddleScale > MIN_SCALE)
            paddleScale--;
    }
    public void setPower(int power)
    {
        currentPower = power;
    }
    
    private final String[][] MAPS = {
//         { "_____0" },
        // Map 0
        { "000000",
          "000000",
          "000000",
          "______",
          "______",
          "______",
          "______" },
        
        // Map 1
        { "__0000",
          "0__000",
          "00__00",
          "000__0",
          "0000__",
          "000__0",
          "00__00" },

        // Map 2
        { "______",
          "000000",
          "000000",
          "__11__",
          "______",
          "000000",
          "_1__1_" },

        // Map 3
        { "000000",
          "000000",
          "110011",
          "______",
          "______",
          "_0000_",
          "__11__" },

        // Map 4
        { "000000",
          "000000",
          "222211",
          "000000",
          "000000",
          "112222",
          "______" },
          
        // Map 5
        { "______",
          "__22__",
          "_2002_",
          "_2002_",
          "_2002_",
          "__22__",
          "______",
          "______",
          "1____1" },

        // Map 6
        { "000000",
          "___33_",
          "000000",
          "_33___",
          "000000",
          "___33_",
          "000000" },

        // Map 7
        { "___333",
          "222000",
          "333___",
          "000222",
          "___333",
          "222000",
          "333___" }
    };
    
    class Ball 
    {
        private double x, y, dx, dy;
        private Double attached;
    
        public Ball()
        {
            attached = .5;
            x = paddle + attached * calculatePaddleGridSize();
            y = SIZE * GRID_BLOCK_SIDE_LENGTH - 1;
            dx = Math.cos(Math.PI / 3) * speed;
            dy = Math.sin(Math.PI / 3) * speed;
        }
        public Ball(double x, double y, double angle)
        {
            this.x = x;
            this.y = y;
            dx = Math.cos(angle) * speed;
            dy = Math.sin(angle) * speed;
        }
        public boolean act()
        {
            if (attached == null)
            {
                int divide = speed;
                for (int i = 0; i < divide; i++)
                {
                    x += dx / divide;
                    if (x < BALL_SIZE / 2)
                    {
                        x = BALL_SIZE / 2;
                        dx *= -1;
                        BOUNCE_SOUND.play();
                    }
                    else if (x > SIZE * GRID_BLOCK_SIDE_LENGTH - BALL_SIZE / 2)
                    {
                        x = SIZE * GRID_BLOCK_SIDE_LENGTH - BALL_SIZE / 2;
                        dx *= -1;
                        BOUNCE_SOUND.play();
                    }
                    y += dy / divide;
                    if (y < BALL_SIZE / 2)
                    {
                        y = BALL_SIZE / 2;
                        dy *= -1;
                        BOUNCE_SOUND.play();
                    }
                    
                    if (dy > 0 && y > SIZE * GRID_BLOCK_SIDE_LENGTH)
                        if (x > paddle && x < paddle + calculatePaddleGridSize())
                        {
                            double d = x - paddle;
                            if (currentPower == Power.REVERSE_POWER)
                                d = calculatePaddleGridSize() - d;
                            double angle = Math.PI + (Math.PI / 8 + d / calculatePaddleGridSize() * Math.PI * 6 / 8);
                            dx = Math.cos(angle) * speed;
                            dy = Math.sin(angle) * speed;
                            if (currentPower == Power.MAGNET_POWER)
                                attached = (x - paddle) / calculatePaddleGridSize();
                            else
                                BOUNCE_SOUND.play();
                        }
                        else if (barrier)
                        {
                            dy *= -1;
                            barrier = false;
                        }
                        else
                            return true;
                    
                    double b = y - (dy / dx) * x;
                    checkDX(b);
                    checkDY(b);
                }
            }
            else
            {
                x = paddle + attached * calculatePaddleGridSize();
                y = SIZE * GRID_BLOCK_SIDE_LENGTH - 1;
            }
            return false;
        }
        private int calculatePaddleGridSize()
        {
            return GRID_BLOCK_SIDE_LENGTH * paddleScale;
        }
        
        private void checkDX(double b)
        {
            int look_offset_x = dx > 0 ? 1 : -1;
            int index_x = (int)x / GRID_BLOCK_SIDE_LENGTH + look_offset_x;
            if (index_x > -1 && index_x < SIZE)
                for (int i = 0; i < 2; i += 2)
                {
                    int index_y = (int)y / GRID_BLOCK_SIDE_LENGTH + i;
                    if (index_y > -1 && index_y < SIZE)
                    {
                        int index_x_2 = (int)x / GRID_BLOCK_SIDE_LENGTH;
                        if (index_x_2 > -1 && index_x_2 < SIZE)
                            if (GRID[index_y][index_x_2].isDestroyed())
                            {
                                int intersectX = (index_x + (look_offset_x == -1 ? 1 : 0)) * GRID_BLOCK_SIDE_LENGTH;
                                int intersectY = (int)((dy / dx) * intersectX + b);
                                if (intersectY >= index_y * GRID_BLOCK_SIDE_LENGTH && intersectY <= (index_y + 1) * GRID_BLOCK_SIDE_LENGTH)
                                    if (Math.sqrt(Math.pow(intersectX - x, 2) + Math.pow(intersectY - y, 2)) < BALL_SIZE / 2)
                                    {
                                        Block block = GRID[index_y][index_x];
                                        if (!block.isDestroyed())
                                        {
                                            hitBlock(block, index_x, index_y);
                                            if (block instanceof SteelBlock || currentPower != Power.SUPER_POWER)
                                                dx *= -1;
                                            return;
                                        }
                                    }
                            }
                    }
                    if (i == 0)
                        i = -3;
                }
        }
        private void checkDY(double b)
        {
            int look_offset_y = dy > 0 ? 1 : -1;
            int index_y = (int)y / GRID_BLOCK_SIDE_LENGTH + look_offset_y;
            if (index_y > -1 && index_y < SIZE)
                for (int i = 0; i < 2; i += 2)
                {
                    int index_x = (int)x / GRID_BLOCK_SIDE_LENGTH + i;
                    if (index_x > -1 && index_x < SIZE)
                    {
                        int index_y_2 = (int)y / GRID_BLOCK_SIDE_LENGTH;
                        if (index_y_2 > -1 && index_y_2 < SIZE)
                            if (GRID[index_y_2][index_x].isDestroyed())
                            {
                                int intersectY = (index_y + (look_offset_y == -1 ? 1 : 0)) * GRID_BLOCK_SIDE_LENGTH;
                                int intersectX = (int)((intersectY - b) / (dy / dx));
                                if (intersectX >= index_x * GRID_BLOCK_SIDE_LENGTH && intersectX <= (index_x + 1) * GRID_BLOCK_SIDE_LENGTH)
                                    if (Math.sqrt(Math.pow(intersectX - x, 2) + Math.pow(intersectY - y, 2)) < BALL_SIZE / 2)
                                    {
                                        Block block = GRID[index_y][index_x];
                                        if (!block.isDestroyed())
                                        {
                                            hitBlock(block, index_x, index_y);
                                            if (block instanceof SteelBlock || currentPower != Power.SUPER_POWER)
                                                dy *= -1;
                                            return;
                                        }
                                    }
                            }
                    }
                    if (i == 0)
                        i = -3;
                }
        }
        private void hitBlock(Block block, int index_x, int index_y)
        {
            block.getHit();
            GRID[index_y][index_x + (index_x % 2 == 1 ? -1 : 1)].getHit();
            if (block.isDestroyed())
                if (--blockCount == 0)
                {
                    if (level == 7 && CodeGenerator.haveUser())
                    {
                        removeObjects(getObjects(CodeDisplay.class));
                        CodeDisplay cd = new CodeDisplay(CODE_ID);
                        addObject(cd, cd.getImage().getWidth() / 2, 6);
                    }
                    level = (level + 1) % MAPS.length;
                    newMap(level);
                    FINISH_SOUND.play();
                }
                else if (Pickup.rollForDrop())
                    PICKUPS.add(Pickup.newPickup((int)((index_x + (index_x % 2 == 0 ? 1 : 0)) * GRID_BLOCK_SIDE_LENGTH), (int)((index_y + .5) * GRID_BLOCK_SIDE_LENGTH)));
        }
        public double getX()
        {
            return x;
        }
        public double getY()
        {
            return y;
        }
        public void updateSpeed()
        {
            double angle = getAngle();
            dx = Math.cos(angle) * speed;
            dy = Math.sin(angle) * speed;
        }
        private double getAngle()
        {
            return Math.atan2(dy, dx);
        }
        public void release()
        {
            attached = null;
        }
        public void split()
        {
            release();
            for (int i = -1; i < 3 && BALLS.size() < 3; i += 2)
                BALLS.add(new Ball(x, y, getAngle() + Math.PI / 12 * i));
        }
    }
}